#!/usr/bin/perl -w
use strict;
use IO::Socket::INET;
use Event qw(loop);
use POSIX qw(EINTR);

package Connection;
use POSIX qw(EINTR EWOULDBLOCK);

sub new {
    my ($class, $socket) = @_;
    $socket->blocking(0);
    my $connection = bless {
        state	   => undef,
        handle	   => $socket,
        rd_watcher => undef,
        wr_watcher => undef,
        in_buffer  => "",
        out_buffer => "",
        nums	   => [],
    }, $class;
    $connection->{rd_watcher} = Event->io(fd => $connection->{handle},
                                          poll => "r",
                                          cb => [$connection, "readable"]);
    $connection->want_num1;
    return $connection;
}

sub readable {
    my $connection = shift;

    my $rc = sysread($connection->{handle}, my $buffer, 8192);
    if ($rc) {
        $connection->{in_buffer} .= $buffer;
        $connection->{state}->($connection, $1) while
            $connection->{in_buffer} =~ s/^(.*)\n//;
        return;
    }
    if (defined($rc)) {
        # EOF
        $connection->drop;
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error reading from socket $!\n";
    $connection->drop;
}

sub writable {
    my $connection = shift;

    if (defined(my $rc = syswrite($connection->{handle},
                                  $connection->{out_buffer}))) {
        substr($connection->{out_buffer}, 0, $rc, "");
        if ($connection->{out_buffer} eq "") {
            $connection->{wr_watcher}->cancel;
            $connection->{wr_watcher} = undef;
        }
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error writing to socket: $!\n";
    $connection->drop;
}

sub drop {
    my $connection = shift;

    if ($connection->{rd_watcher}) {
        $connection->{rd_watcher}->cancel;
        $connection->{rd_watcher} = undef;
    }
    if ($connection->{wr_watcher}) {
        $connection->{wr_watcher}->cancel;
        $connection->{wr_watcher} = undef;
    }
    close $connection->{handle};
}

sub write : method {
    my $connection = shift;
    $connection->{out_buffer} .= shift;
    return if $connection->{out_buffer} eq "" || $connection->{wr_watcher};
    $connection->{wr_watcher} = Event->io(fd => $connection->{handle},
                                          poll => "w",
                                          cb => [$connection, "writable"]);
}

sub want_num1 {
    my $connection = shift;
    $connection->write("First Number ?\n");
    $connection->{state} = \&got_num1;
}

sub got_num1 {
    my $connection = shift;
    $connection->{nums}[0] = shift;
    $connection->want_num2;
}

sub want_num2 {
    my $connection = shift;
    $connection->write("Second Number ?\n");
    $connection->{state} =  \&got_num2;
}

sub got_num2 {
    my $connection = shift;
    my $num1 = $connection->{nums}[0];
    my $num2 = shift;
    $connection->write(sprintf("%d*%d=%d\n", $num1, $num2, $num1*$num2));
    $connection->want_num1;
}

sub DESTROY {
    my $connection = shift;
    print STDERR "Connection forgotten\n";
}

package main;

$SIG{PIPE} = "IGNORE";

my $listener = IO::Socket::INET->new(LocalAddr => "127.0.0.1",
                                     LocalPort => 1234,
                                     Listen    => 5) ||
    die "Could not listen on 127.0.0.1:1234: $!\n";

$listener->blocking(0);
Event->io(fd => $listener,
          poll => "r",
          cb => sub {
              if (my $socket = $listener->accept) {
                  Connection->new($socket);
              } else {
                  print STDERR "Error accepting connection: $!\n" unless
                      $! == EINTR;
              }
          });

loop;
