#!/usr/bin/perl -w
use strict;
use IO::Socket::INET;
use IO::Select;
use POSIX qw(EINTR);

my $rd_set = IO::Select->new();
my $wr_set = IO::Select->new();

my %connections;

package Connection;
use POSIX qw(EINTR EWOULDBLOCK);

sub new {
    my ($class, $socket, $out) = @_;
    $socket->blocking(0);
    my $connection = bless {
        handle	   => $socket,
        reading	   => 1,
        in_buffer  => "",
        out_buffer => $out,
        peer_ip    => $socket->peerhost,
        peer_port  => $socket->peerport,
    }, $class;
    $rd_set->add($connection->{handle});
    $wr_set->add($connection->{handle}) if $out ne "";
    print STDERR "Handling connection from $connection->{peer_ip}:$connection->{peer_port}\n";
    return $connection;
}

sub readable {
    my $connection = shift;

    my $rc = sysread($connection->{handle}, my $buffer, 8192);
    if ($rc) {
        $connection->{in_buffer} .= $buffer;
        return;
    }
    if (defined($rc)) {
        # EOF
        printf STDERR "Read %d bytes\n", length($connection->{in_buffer});
        $connection->{reading} = 0;
        $rd_set->remove($connection->{handle});
        shutdown($connection->{handle}, 0);
        $connection->remove if $connection->{out_buffer} eq "";
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error reading from socket to $connection->{peer_ip}:$connection->{peer_port}: $!\n";
    $connection->drop;
}

sub writable {
    my $connection = shift;

    if (defined(my $rc = syswrite($connection->{handle},
                                  $connection->{out_buffer}))) {
        substr($connection->{out_buffer}, 0, $rc, "");
        if ($connection->{out_buffer} eq "") {
            $wr_set->remove($connection->{handle});
            shutdown($connection->{handle}, 1);
            $connection->remove if !$connection->{reading};
        }
        return;
    }
    return if $! == EINTR || $! == EWOULDBLOCK;
    print STDERR "Error writing to socket to $connection->{peer_ip}:$connection->{peer_port}: $!\n";
    $connection->drop;
}

sub remove {
    my $connection = shift;
    delete $connections{$connection->{handle}};
}

sub drop {
    my $connection = shift;
    $connection->remove;
    $rd_set->remove($connection->{handle});
    $wr_set->remove($connection->{handle});
    $connection->{reading} = 0;
    close $connection->{handle};
}

sub DESTROY {
    my $connection = shift;
    print STDERR "Connection to $connection->{peer_ip}:$connection->{peer_port} forgotten\n";
}

package main;

$SIG{PIPE} = "IGNORE";

defined(my $file = shift) || die "Usage: $0 file\n";
open(my $fh, "<", $file) || die "Could not open $file: $!";
my $out = do { local $/; <$fh> };

my $listener = IO::Socket::INET->new(LocalAddr => "127.0.0.1",
                                     LocalPort => 1234,
                                     Listen    => 5) ||
    die "Could not listen on 127.0.0.1:1234: $!\n";

$listener->blocking(0);
$rd_set->add($listener);

while (1) {
    $! = EINTR;
    if (my ($readable, $writable) =
        IO::Select::select($rd_set, $wr_set, undef, undef)) {
        $connections{$_}->writable for @$writable;
        for (@$readable) {
            if ($_ eq $listener) {
                if (my $socket = $listener->accept) {
                    my $connection = Connection->new($socket, $out);
                    $connections{$socket} = $connection;
                } else {
                    print STDERR "Error accepting connection: $!\n" unless
                        $! == EINTR;
                }
            } else {
                $connections{$_}->readable if $connections{$_}{reading};
            }
        }
    } else {
        die "Select failed: $!" unless $! == EINTR;
    }
}
